@@ -1,43 +1,36 @@
-module EVM.Test.BlockchainTests where
+module EVM.Test.BlockchainTests (prepareTests, problematicTests, Case, vmForCase, checkExpectation, allTestCases) where
 
 import EVM (initialContract, makeVm, setEIP4788Storage)
 import EVM.Concrete qualified as EVM
+import EVM.Effects
+import EVM.Expr (maybeLitAddrSimp)
 import EVM.FeeSchedule (feeSchedule)
 import EVM.Fetch qualified
-import EVM.Format (hexText)
+import EVM.Solvers (withSolvers, Solver(..))
 import EVM.Stepper qualified
 import EVM.Transaction
-import EVM.UnitTest (writeTrace)
 import EVM.Types hiding (Block, Case, Env)
-import EVM.Expr (maybeLitWordSimp, maybeLitAddrSimp)
-import EVM.Tracing qualified as Tracing
-import EVM.Test.FuzzSymExec (compareTraces, EVMToolTraceOutput(..), decodeTraceOutputHelper)
+import EVM.UnitTest (writeTrace)
 
 import Optics.Core
 import Control.Arrow ((***), (&&&))
 import Control.Monad
-import Control.Monad.ST (RealWorld, stToIO)
+import Control.Monad.ST (stToIO)
 import Control.Monad.State.Strict
 import Control.Monad.IO.Unlift
-import EVM.Effects
 import Data.Aeson ((.:), (.:?), FromJSON (..))
 import Data.Aeson qualified as JSON
 import Data.Aeson.Types qualified as JSON
 import Data.ByteString qualified as BS
 import Data.ByteString.Lazy qualified as Lazy
-import Data.ByteString.Lazy qualified as LazyByteString
-import Data.List (isInfixOf, isPrefixOf)
 import Data.Map (Map)
 import Data.Map qualified as Map
 import Data.Maybe (fromJust, fromMaybe, isNothing, isJust)
 import Data.Word (Word64)
-import System.Environment (lookupEnv, getEnv)
+import GHC.Generics (Generic)
+import System.Environment (getEnv)
 import System.FilePath.Find qualified as Find
 import System.FilePath.Posix (makeRelative, (</>))
-import System.IO (hPutStr, hClose)
-import System.IO.Temp (withSystemTempFile)
-import System.Process (readProcessWithExitCode)
-import GHC.IO.Exception (ExitCode(ExitSuccess))
 import Witch (into, unsafeInto)
 import Witherable (Filterable, catMaybes)
 
@@ -59,70 +52,96 @@ data Block = Block
   , beaconRoot  :: W256
   } deriving Show
 
+data BlockchainContract = BlockchainContract
+  { code    :: ByteStringS
+  , nonce   :: W64
+  , balance :: W256
+  , storage :: Map W256 W256
+  } deriving (Eq, Show, Generic)
+
+instance FromJSON BlockchainContract
+
+asBCContract :: Contract -> BlockchainContract
+asBCContract c = BlockchainContract code nonce balance storage
+  where
+    code = case c.code of
+      (RuntimeCode (ConcreteRuntimeCode bs)) -> ByteStringS bs
+      _ -> internalError "Expected concrete contract"
+    nonce = fromJust c.nonce
+    balance = forceLit (c.balance)
+    storage = fromConcrete c.storage
+
+makeContract :: BlockchainContract -> Contract
+makeContract (BlockchainContract (ByteStringS code) nonce balance storage) =
+    initialContract (RuntimeCode (ConcreteRuntimeCode code))
+      & set #nonce    (Just nonce)
+      & set #balance  (Lit balance)
+      & set #storage (ConcreteStore storage)
+      & set #origStorage (ConcreteStore storage)
+
+type BlockchainContracts = Map Addr BlockchainContract
+
 data Case = Case
   { vmOpts      :: VMOpts Concrete
-  , checkContracts  :: Map Addr Contract
-  , testExpectation :: Map Addr Contract
+  , checkContracts  :: BlockchainContracts
+  , testExpectation :: BlockchainContracts
   } deriving Show
 
 data BlockchainCase = BlockchainCase
   { blocks  :: [Block]
-  , pre     :: Map Addr Contract
-  , post    :: Map Addr Contract
+  , pre     :: BlockchainContracts
+  , post    :: BlockchainContracts
   , network :: String
   } deriving Show
 
-
-testEnv :: Env
-testEnv = Env { config = defaultConfig }
-
-main :: IO ()
-main = do
-  tests <- runEnv testEnv prepareTests
-  defaultMain tests
-
 prepareTests :: App m => m TestTree
 prepareTests = do
-  repo <- liftIO $ getEnv "HEVM_ETHEREUM_TESTS_REPO"
-  let testsDir = "BlockchainTests/GeneralStateTests"
-  let dir = repo </> testsDir
-  jsonFiles <- liftIO $ Find.find Find.always (Find.extension Find.==? ".json") dir
-  liftIO $ putStrLn $ "Loading and parsing json files from ethereum-tests from " <> show dir
-  isCI <- liftIO $ isJust <$> lookupEnv "CI"
-  let problematicTests = if isCI then commonProblematicTests <> ciProblematicTests else commonProblematicTests
-  let ignoredFiles = if isCI then ciIgnoredFiles else []
-  groups <- mapM (\f -> testGroup (makeRelative repo f) <$> (if any (`isInfixOf` f) ignoredFiles then pure [] else testsFromFile f problematicTests)) jsonFiles
+  rootDir <- liftIO rootDirectory
+  liftIO $ putStrLn $ "Loading and parsing json files from ethereum-tests from " <> show rootDir
+  cases <- liftIO allTestCases
+  groups <- forM (Map.toList cases) (\(f, subtests) -> testGroup (makeRelative rootDir f) <$> (process subtests))
   liftIO $ putStrLn "Loaded."
   pure $ testGroup "ethereum-tests" groups
-
-testsFromFile
-  :: forall m . App m
-  => String -> Map String (TestTree -> TestTree) -> m [TestTree]
-testsFromFile fname problematicTests = do
-  fContents <- liftIO $ LazyByteString.readFile fname
-  let parsed = parseBCSuite fContents
-  liftIO $ putStrLn $ "Parsed " <> fname
-  case parsed of
-    Left "No cases to check." -> pure []
-    Left _err -> pure []
-    Right allTests -> mapM runTest $ Map.toList allTests
   where
-    runTest :: (String, Case) -> m TestTree
+    process :: forall m . App m => (Map String Case) -> m [TestTree]
+    process tests = forM (Map.toList tests) runTest
+
+    runTest :: App m => (String, Case) -> m TestTree
     runTest (name, x) = do
-      exec <- toIO $ runVMTest x
+      let fetcher q = withSolvers Z3 0 1 (Just 0) $ \s -> EVM.Fetch.noRpcFetcher s q
+      exec <- toIO $ runVMTest fetcher x
       pure $ testCase' name exec
     testCase' :: String -> Assertion -> TestTree
     testCase' name assertion =
       case Map.lookup name problematicTests of
-        Just f -> f (testCase name (liftIO assertion))
-        Nothing -> testCase name (liftIO assertion)
+        Just f -> f (testCase name assertion)
+        Nothing -> testCase name assertion
 
--- CI has issues with some heaver tests, disable in bulk
-ciIgnoredFiles :: [String]
-ciIgnoredFiles = []
+rootDirectory :: IO FilePath
+rootDirectory = do
+  repo <- getEnv "HEVM_ETHEREUM_TESTS_REPO"
+  let testsDir = "BlockchainTests/GeneralStateTests"
+  pure $ repo </> testsDir
 
-commonProblematicTests :: Map String (TestTree -> TestTree)
-commonProblematicTests = Map.fromList
+collectJsonFiles :: FilePath -> IO [FilePath]
+collectJsonFiles rootDir = Find.find Find.always (Find.extension Find.==? ".json") rootDir
+
+allTestCases :: IO (Map FilePath (Map String Case))
+allTestCases = do
+  root <- rootDirectory
+  jsons <- collectJsonFiles root
+  cases <- forM jsons (\fname -> do
+      fContents <- BS.readFile fname
+      let parsed = case (parseBCSuite (Lazy.fromStrict fContents)) of
+                    Left "No cases to check." -> mempty
+                    Left _err -> mempty -- TODO: This should be an error
+                    Right allTests -> allTests
+      pure (fname, parsed)
+    )
+  pure $ Map.fromList cases
+
+problematicTests :: Map String (TestTree -> TestTree)
+problematicTests = Map.fromList
   [ ("loopMul_d0g0v0_Cancun", ignoreTestBecause "hevm is too slow")
   , ("loopMul_d1g0v0_Cancun", ignoreTestBecause "hevm is too slow")
   , ("loopMul_d2g0v0_Cancun", ignoreTestBecause "hevm is too slow")
@@ -159,70 +178,70 @@ commonProblematicTests = Map.fromList
   , ("failed_tx_xcf416c53_d0g0v0_Cancun", ignoreTestBecause "EIP-4844 not implemented")
   ]
 
-ciProblematicTests :: Map String (TestTree -> TestTree)
-ciProblematicTests = Map.fromList
-  [ ("Return50000_d0g1v0_Cancun", ignoreTest)
-  , ("Return50000_2_d0g1v0_Cancun", ignoreTest)
-  , ("randomStatetest177_d0g0v0_Cancun", ignoreTest)
-  , ("static_Call50000_d0g0v0_Cancun", ignoreTest)
-  , ("static_Call50000_d1g0v0_Cancun", ignoreTest)
-  , ("static_Call50000bytesContract50_1_d1g0v0_Cancun", ignoreTest)
-  , ("static_Call50000bytesContract50_2_d1g0v0_Cancun", ignoreTest)
-  , ("static_Return50000_2_d0g0v0_Cancun", ignoreTest)
-  , ("loopExp_d10g0v0_Cancun", ignoreTest)
-  , ("loopExp_d11g0v0_Cancun", ignoreTest)
-  , ("loopExp_d12g0v0_Cancun", ignoreTest)
-  , ("loopExp_d13g0v0_Cancun", ignoreTest)
-  , ("loopExp_d14g0v0_Cancun", ignoreTest)
-  , ("loopExp_d8g0v0_Cancun", ignoreTest)
-  , ("loopExp_d9g0v0_Cancun", ignoreTest)
-  ]
 
-runVMTest :: App m => Case -> m ()
-runVMTest x = do
+runVMTest :: App m => EVM.Fetch.Fetcher Concrete m -> Case -> m ()
+runVMTest fetcher x = do
   -- traceVsGeth fname name x
   vm0 <- liftIO $ vmForCase x
-  result <- EVM.Stepper.interpret (EVM.Fetch.zero 0 (Just 0)) vm0 EVM.Stepper.runFully
+  result <- EVM.Stepper.interpret fetcher vm0 EVM.Stepper.runFully
   writeTrace result
-  maybeReason <- checkExpectation x result
-  liftIO $ forM_ maybeReason assertFailure
-
-
--- | Run a vm test and output a geth style per opcode trace
-traceVMTest :: App m => Case -> m [Tracing.VMTraceStep]
-traceVMTest x = do
-  vm0 <- liftIO $ vmForCase x
-  (_, (_, ts)) <- runStateT (Tracing.interpretWithTrace (EVM.Fetch.zero 0 (Just 0)) EVM.Stepper.runFully) (vm0, [])
-  pure ts
-
--- | given a path to a test file, a test case from within that file, and a trace from geth from running that test, compare the traces and show where we differ
--- This would need a few tweaks to geth to make this really usable (i.e. evm statetest show allow running a single test from within the test file).
-traceVsGeth :: App m => String -> String -> Case -> m ()
-traceVsGeth fname name x = do
-  liftIO $ putStrLn "-> Running `evm --json blocktest` tool."
-  (exitCode, evmtoolStdout, evmtoolStderr) <- liftIO $ readProcessWithExitCode "evm" [
-                               "--json"
-                               , "blocktest"
-                               , "--run", name
-                               , fname
-                               ] ""
-  when (exitCode /= ExitSuccess) $ liftIO $ do
-    putStrLn $ "evmtool exited with code " <> show exitCode
-    putStrLn $ "evmtool stderr output:" <> show evmtoolStderr
-    putStrLn $ "evmtool stdout output:" <> show evmtoolStdout
-  hevm <- traceVMTest x
-  decodedContents <- liftIO $ withSystemTempFile "trace.jsonl" $ \traceFile hdl -> do
-    hPutStr hdl $ filterInfoLines evmtoolStderr
-    hClose hdl
-    decodeTraceOutputHelper traceFile
-  let EVMToolTraceOutput ts _ = fromJust decodedContents
-  liftIO $ putStrLn "Comparing traces."
-  _ <- liftIO $ compareTraces hevm ts
-  pure ()
+  let maybeReason = checkExpectation x result
+  liftIO $ forM_ maybeReason (liftIO >=> assertFailure)
 
+checkExpectation :: Case -> VM Concrete -> Maybe (IO String)
+checkExpectation x vm = let (okState, okBal, okNonce, okStor, okCode) = checkExpectedContracts vm x.testExpectation in
+  if okState then Nothing else Just $ checkStateFail x (okBal, okNonce, okStor, okCode)
   where
-    filterInfoLines :: String -> String
-    filterInfoLines input = unlines $ filter (not . isPrefixOf "INFO") (lines input)
+    checkExpectedContracts :: VM Concrete -> BlockchainContracts -> (Bool, Bool, Bool, Bool, Bool)
+    checkExpectedContracts vm' expected =
+      let cs = fmap (asBCContract . clearZeroStorage) $ forceConcreteAddrs vm'.env.contracts
+      in ( (expected ~= cs)
+        , (clearBalance <$> expected) ~= (clearBalance <$> cs)
+        , (clearNonce   <$> expected) ~= (clearNonce   <$> cs)
+        , (clearStorage <$> expected) ~= (clearStorage <$> cs)
+        , (clearCode    <$> expected) ~= (clearCode    <$> cs)
+        )
+
+    -- quotient account state by nullness
+    (~=) :: BlockchainContracts -> BlockchainContracts -> Bool
+    (~=) cs1 cs2 =
+        let nullAccount = asBCContract $ EVM.initialContract (RuntimeCode (ConcreteRuntimeCode ""))
+            padNewAccounts cs ks = Map.union cs $ Map.fromList [(k, nullAccount) | k <- ks]
+            padded_cs1 = padNewAccounts cs1 (Map.keys cs2)
+            padded_cs2 = padNewAccounts cs2 (Map.keys cs1)
+        in and $ zipWith (==) (Map.elems padded_cs1) (Map.elems padded_cs2)
+    
+    checkStateFail :: Case -> (Bool, Bool, Bool, Bool) -> IO String
+    checkStateFail x' (okBal, okNonce, okData, okCode) = do
+      let
+        printContracts :: BlockchainContracts -> IO ()
+        printContracts cs = putStrLn $ Map.foldrWithKey (\k c acc ->
+          acc ++ "-->" <> show k ++ " : "
+                      ++ (show c.nonce) ++ " "
+                      ++ (show c.balance) ++ " "
+                      ++ (show c.storage)
+            ++ "\n") "" cs
+
+        reason = map fst (filter (not . snd)
+            [ ("bad-state",       okBal   || okNonce || okData  || okCode)
+            , ("bad-balance", not okBal   || okNonce || okData  || okCode)
+            , ("bad-nonce",   not okNonce || okBal   || okData  || okCode)
+            , ("bad-storage", not okData  || okBal   || okNonce || okCode)
+            , ("bad-code",    not okCode  || okBal   || okNonce || okData)
+            ])
+        check = x'.checkContracts
+        expected = x'.testExpectation
+        actual = fmap (asBCContract . clearZeroStorage) $ forceConcreteAddrs vm.env.contracts
+
+      putStrLn $ "-> Failing because of: " <> (unwords reason)
+      putStrLn "-> Pre balance/state: "
+      printContracts check
+      putStrLn "-> Expected balance/state: "
+      printContracts expected
+      putStrLn "-> Actual balance/state: "
+      printContracts actual
+      pure (unwords reason)
+
 
 splitEithers :: (Filterable f) => f (Either a b) -> (f a, f b)
 splitEithers =
@@ -234,115 +253,23 @@ fromConcrete :: Expr Storage -> Map W256 W256
 fromConcrete (ConcreteStore s) = s
 fromConcrete s = internalError $ "unexpected abstract store: " <> show s
 
-checkStateFail :: Case -> VM Concrete RealWorld -> (Bool, Bool, Bool, Bool) -> IO String
-checkStateFail x vm (okBal, okNonce, okData, okCode) = do
-  let
-    printContracts :: Map Addr Contract -> IO ()
-    printContracts cs = putStrLn $ Map.foldrWithKey (\k c acc ->
-      acc ++ "-->" <> show k ++ " : "
-                   ++ (show $ fromJust c.nonce) ++ " "
-                   ++ (show $ fromJust $ maybeLitWordSimp c.balance) ++ " "
-                   ++ (show $ fromConcrete c.storage)
-                   ++ (show $ fromConcrete c.tStorage)
-        ++ "\n") "" cs
-
-    reason = map fst (filter (not . snd)
-        [ ("bad-state",       okBal   || okNonce || okData  || okCode)
-        , ("bad-balance", not okBal   || okNonce || okData  || okCode)
-        , ("bad-nonce",   not okNonce || okBal   || okData  || okCode)
-        , ("bad-storage", not okData  || okBal   || okNonce || okCode)
-        , ("bad-code",    not okCode  || okBal   || okNonce || okData)
-        ])
-    check = x.checkContracts
-    expected = x.testExpectation
-    actual = fmap (clearZeroStorage . clearOrigStorage) $ forceConcreteAddrs vm.env.contracts
-
-  putStrLn $ "-> Failing because of: " <> (unwords reason)
-  putStrLn "-> Pre balance/state: "
-  printContracts check
-  putStrLn "-> Expected balance/state: "
-  printContracts expected
-  putStrLn "-> Actual balance/state: "
-  printContracts actual
-  pure (unwords reason)
-
-checkExpectation
-  :: App m
-  => Case -> VM Concrete RealWorld -> m (Maybe String)
-checkExpectation x vm = do
-  let expectation = x.testExpectation
-      (okState, okBal, okNonce, okStor, okCode) = checkExpectedContracts vm expectation
-  if okState then do
-    pure Nothing
-  else liftIO $ Just <$> checkStateFail x vm (okBal, okNonce, okStor, okCode)
-
--- quotient account state by nullness
-(~=) :: Map Addr Contract -> Map Addr Contract -> Bool
-(~=) cs1 cs2 =
-    let nullAccount = EVM.initialContract (RuntimeCode (ConcreteRuntimeCode ""))
-        padNewAccounts cs ks = Map.union cs $ Map.fromList [(k, nullAccount) | k <- ks]
-        padded_cs1 = padNewAccounts cs1 (Map.keys cs2)
-        padded_cs2 = padNewAccounts cs2 (Map.keys cs1)
-    in and $ zipWith (===) (Map.elems padded_cs1) (Map.elems padded_cs2)
-
-(===) :: Contract -> Contract -> Bool
-c1 === c2 =
-  codeEqual && storageEqual && c1.balance == c2.balance && c1.nonce == c2.nonce
-  where
-    storageEqual = c1.storage == c2.storage
-    codeEqual = case (c1.code, c2.code) of
-      (RuntimeCode a', RuntimeCode b') -> a' == b'
-      codes -> internalError ("unexpected code: \n" <> show codes)
-
-checkExpectedContracts :: VM Concrete RealWorld -> Map Addr Contract -> (Bool, Bool, Bool, Bool, Bool)
-checkExpectedContracts vm expected =
-  let cs = fmap (clearZeroStorage . clearOrigStorage) $ forceConcreteAddrs vm.env.contracts
-  in ( (expected ~= cs)
-     , (clearBalance <$> expected) ~= (clearBalance <$> cs)
-     , (clearNonce   <$> expected) ~= (clearNonce   <$> cs)
-     , (clearStorage <$> expected) ~= (clearStorage <$> cs)
-     , (clearCode    <$> expected) ~= (clearCode    <$> cs)
-     )
-
-clearOrigStorage :: Contract -> Contract
-clearOrigStorage = set #origStorage (ConcreteStore mempty)
-
 clearZeroStorage :: Contract -> Contract
 clearZeroStorage c = case c.storage of
   ConcreteStore m -> let store = Map.filter (/= 0) m
                      in set #storage (ConcreteStore store) c
   _ -> internalError "Internal Error: unexpected abstract store"
 
-clearStorage :: Contract -> Contract
-clearStorage c = c { storage = clear c.storage, tStorage = clear c.tStorage }
-  where
-    clear :: Expr Storage -> Expr Storage
-    clear (ConcreteStore _) = ConcreteStore mempty
-    clear _ = internalError "Internal Error: unexpected abstract store"
+clearStorage :: BlockchainContract -> BlockchainContract
+clearStorage c = c { storage = mempty}
 
-clearBalance :: Contract -> Contract
-clearBalance c = set #balance (Lit 0) c
+clearBalance :: BlockchainContract -> BlockchainContract
+clearBalance c = c {balance = 0}
 
-clearNonce :: Contract -> Contract
-clearNonce c = set #nonce (Just 0) c
+clearNonce :: BlockchainContract -> BlockchainContract
+clearNonce c = c {nonce = 0}
 
-clearCode :: Contract -> Contract
-clearCode c = set #code (RuntimeCode (ConcreteRuntimeCode "")) c
-
-instance FromJSON Contract where
-  parseJSON (JSON.Object v) = do
-    code <- (RuntimeCode . ConcreteRuntimeCode <$> (hexText <$> v .: "code"))
-    storage <- v .: "storage"
-    balance <- v .: "balance"
-    nonce   <- v .: "nonce"
-    pure $ EVM.initialContract code
-             & #balance .~ (Lit balance)
-             & #nonce   ?~ nonce
-             & #storage .~ (ConcreteStore storage)
-             & #origStorage .~ (ConcreteStore storage)
-
-  parseJSON invalid =
-    JSON.typeMismatch "Contract" invalid
+clearCode :: BlockchainContract -> BlockchainContract
+clearCode c = c {code = (ByteStringS "")}
 
 instance FromJSON BlockchainCase where
   parseJSON (JSON.Object v) = BlockchainCase
@@ -372,7 +299,7 @@ instance FromJSON Block where
   parseJSON invalid =
     JSON.typeMismatch "Block" invalid
 
-parseContracts :: Which -> JSON.Object -> JSON.Parser (Map Addr Contract)
+parseContracts :: Which -> JSON.Object -> JSON.Parser (BlockchainContracts)
 parseContracts w v = v .: which >>= parseJSON
   where which = case w of
           Pre  -> "pre"
@@ -424,7 +351,7 @@ maxCodeSize :: W256
 maxCodeSize = 24576
 
 fromBlockchainCase' :: Block -> Transaction
-                       -> Map Addr Contract -> Map Addr Contract
+                       -> BlockchainContracts -> BlockchainContracts
                        -> Either BlockchainError Case
 fromBlockchainCase' block tx preState postState =
   let isCreate = isNothing tx.toAddr in
@@ -464,11 +391,13 @@ fromBlockchainCase' block tx preState postState =
       postState
         where
           toAddr = maybe (EVM.createAddress origin (fromJust senderNonce)) LitAddr (tx.toAddr)
-          senderNonce = view (accountAt (LitAddr origin) % #nonce) (Map.mapKeys LitAddr preState)
+          senderNonce = (.nonce) <$> Map.lookup origin preState
           toCode = Map.lookup toAddr (Map.mapKeys LitAddr preState)
           theCode = if isCreate
                     then InitCode tx.txdata mempty
-                    else maybe (RuntimeCode (ConcreteRuntimeCode "")) (.code) toCode
+                    else RuntimeCode . ConcreteRuntimeCode $ case toCode of
+                      Nothing ->  ""
+                      Just (BlockchainContract (ByteStringS bs) _ _ _) -> bs
           effectiveGasPrice = effectiveprice tx block.baseFee
           cd = if isCreate
                then mempty
@@ -495,43 +424,41 @@ maxBaseFee tx =
      EIP1559Transaction -> fromJust tx.maxFeePerGas
      _ -> fromJust tx.gasPrice
 
-checkTx :: Transaction -> Block -> Map Addr Contract -> Maybe (Map Addr Contract)
+checkTx :: Transaction -> Block -> BlockchainContracts -> Maybe (BlockchainContracts)
 checkTx tx block prestate = do
   origin <- sender tx
   validateTx tx block prestate
-  let isCreate    = isNothing tx.toAddr
-      cs          = Map.mapKeys LitAddr prestate
-      senderNonce = view (accountAt (LitAddr origin) % #nonce) cs
-      toAddr      = maybe (EVM.createAddress origin (fromJust senderNonce)) LitAddr (tx.toAddr)
-      prevCode    = view (accountAt toAddr % #code) cs
-      prevNonce   = view (accountAt toAddr % #nonce) cs
-
-      nonEmptyAccount = case prevCode of
-                        RuntimeCode (ConcreteRuntimeCode b) -> not (BS.null b)
-                        _ -> True
-      badNonce = prevNonce /= Just 0
-      initCodeSizeExceeded = BS.length tx.txdata > (unsafeInto maxCodeSize * 2)
-  if isCreate && (badNonce || nonEmptyAccount || initCodeSizeExceeded)
-  then mzero
+  if (isJust tx.toAddr) then pure prestate
   else
-    pure prestate
+    let senderNonce = (.nonce) <$> Map.lookup origin prestate
+        addr  = case EVM.createAddress origin (fromJust senderNonce) of
+                  (LitAddr a) -> a
+                  _ -> internalError "Cannot happen"
+        freshContract = BlockchainContract (ByteStringS "") 0 0 mempty
+        (BlockchainContract (ByteStringS b) prevNonce _ _) = (fromMaybe freshContract $ Map.lookup addr prestate)
+        nonEmptyAccount = not (BS.null b)
+        badNonce = prevNonce /= 0
+        initCodeSizeExceeded = BS.length tx.txdata > (unsafeInto maxCodeSize * 2)
+    in
+    if (badNonce || nonEmptyAccount || initCodeSizeExceeded) then mzero
+    else pure prestate
 
-validateTx :: Transaction -> Block -> Map Addr Contract -> Maybe ()
+validateTx :: Transaction -> Block -> BlockchainContracts -> Maybe ()
 validateTx tx block cs = do
   origin <- sender tx
-  Lit originBalance <- (.balance) <$> Map.lookup origin cs
+  originBalance <- (.balance) <$> Map.lookup origin cs
   originNonce <- (.nonce) <$> Map.lookup origin cs
   let gasDeposit = (effectiveprice tx block.baseFee) * (into tx.gasLimit)
   if gasDeposit + tx.value <= originBalance
-    && (Just (unsafeInto tx.nonce) == originNonce) && block.baseFee <= maxBaseFee tx
+    && ((unsafeInto tx.nonce) == originNonce) && block.baseFee <= maxBaseFee tx
   then Just ()
   else Nothing
 
-vmForCase :: Case -> IO (VM Concrete RealWorld)
+vmForCase :: Case -> IO (VM Concrete)
 vmForCase x = do
   vm <- stToIO $ makeVm x.vmOpts
     -- TODO: why do we override contracts here instead of using VMOpts otherContracts?
-    <&> set (#env % #contracts) (Map.mapKeys LitAddr x.checkContracts)
+    <&> set (#env % #contracts) (Map.mapKeys LitAddr $ Map.map makeContract x.checkContracts)
     -- TODO: we need to call this again because we override contracts in the
     -- previous line
     <&> setEIP4788Storage x.vmOpts
